=begin
Copyright 2013-2014, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         ToolbarEditor.rb

Author:       Andreas Eisenbarth

Usage:        Menu 'Window' → 'ToolbarEditor'
              Opens a dialog where you can drag and drop toolbar buttons from the
              list on the right into toolbar panels on the left. Click "Apply" to
              save the changes. You need to restart SketchUp for all changes to
              take effect.

Description:  This plugin lets you create your own toolbars.

Required:     SketchUp 8 M1 or higher

Version:      1.1.0

Date:         05.04.2014


Information:  The principle of custom toolbars is to save&read the defined toolbars
              from a file, then loop over each toolbar configuration and create
              a new UI::Toolbar and add buttons for its commands.
              However there is no standard way in SketchUp to detect features/
              commands of other plugins and execute them (UI::Command.proc.call).
              One way is to write code for each plugin's method name manually in
              a configuration file or script (so the user would need to look 
              into each plugin's source code).
              Another way is to get all UI::Command instances from Ruby's ObjectSpace.
              Unfortunately UI::Command don't have persistent identifiers to
              remember them over SketchUp sessions. The Plugin LaunchUp therefore
              creates an index and generates a unique and persistent hash from
              each command's metadata. This plugin uses LaunchUp's index (if installed)
              or a lightweight substitute.
              Because of that it is important that the interception (in LaunchUp)
              happens before other scripts load, and that custom toolbars are
              created after all other scripts have loaded.
TODO: what happens when a command has no icon? allow adding icon (edit → edit dialog → change icon)
TODO: avoid duplicate buttons
TODO: unique id (Fixnum!)
TODO: css of buttons → smaller and on the right; hide when draggin to the left
=end
require 'sketchup.rb'


module AE


module ToolbarEditor


# This plugin's folder.
DIR = File.dirname(__FILE__) unless defined?(self::DIR)

# Translation library.
require(File.join(DIR, 'Translate.rb'))

# Load translation strings and add plugin to UI.
TRANSLATE = Translate.new("ToolbarEditor", File.join(DIR, "resources")) unless defined?(self::TRANSLATE)

# WebDialog Helper.
require(File.join(DIR, 'Dialog.rb'))

# Index of all commands.
if defined?(AE::LaunchUp) && defined?(AE::LaunchUp::Index) &&
    ( AE::LaunchUp::Index.instance_methods.include?(:get_all) ||
      AE::LaunchUp::Index.instance_methods.include?("get_all") )
  Index = AE::LaunchUp::Index unless defined?(self::Index)
else
  require(File.join(DIR, 'Index.rb'))
end

# Options.
require(File.join(DIR, 'Options.rb'))
@options ||= Options.new("ToolbarEditor", {
  :toolbars => [], # { "name" => [String], "buttons" => [Array<Fixnum>] }
  :snippets => [], # { "id" => [Fixnum], "name" => [String], … }
  #:workspaces => [], # {:name => "workspace1", :toolbars => [array of ids]}
})

# Reference to the main dialog.
@editor_dlg ||= nil
# Reference for toolbar objects: :id => UI::Toolbar
@toolbars ||= {}
# Reference for custom commands objects: :id => UI::Command
@snippets ||= []

# Data directory for saving files.
dir = [
       File.join(DIR, "data"),                         # Plugins folder
       ENV["APPDATA"],                                 # Windows
       File.join(ENV["HOME"].to_s, ".local", "share"), # Free desktop standard
       File.join(ENV["HOME"].to_s, "Library", "Application Support"), # OS X
       File.expand_path(".")                           # Fallback: user's folder.
      ].compact.find{ |path| File.exists?(path) && File.writable?(path) }
DATA_DIR = File.join(File.expand_path(dir), "SketchUp ToolbarEditor") unless defined?(self::DATA_DIR)


public


# This displays the main dialog with search field.
def self.show_dialog
  # If the dialog exists, bring it to the front.
  if @editor_dlg && @editor_dlg.visible?
    @editor_dlg.bring_to_front
  else
    @editor_dlg = AE::ToolbarEditor::Dialog.new({
      :dialog_title => TRANSLATE["Toolbar Editor"],
      :scrollable => false,
      :preferences_key => "AE_ToolbarEditor",
      :height => 550,
      :width => 400,
      :left => 600,
      :top => 200,
      :resizable => true
    })
    @editor_dlg.min_width = 420
    @editor_dlg.min_height = 350
    html_path = File.join(DIR, "html", "ToolbarEditor.html")
    @editor_dlg.set_file(html_path)

    @editor_dlg.on_show { |dlg|
      TRANSLATE.webdialog(dlg)
      # Wait a short moment so that the loading animation can start before the dialog is busy loading the index.
      UI.start_timer(0.1, false){
        buttons = Index.instance.get_all{ |hash|
          hash[:icon].is_a?(String) && !hash[:icon].empty?
        }.sort_by{ |entry|
          entry[:name] || ""
        }.sort_by{ |entry|
          entry[:category] || ""
        }
        dlg.call_function("AE.ToolbarEditor.loadCustomButtons", @snippets)
        dlg.call_function("AE.ToolbarEditor.loadButtons", buttons)
        dlg.call_function("AE.ToolbarEditor.loadToolbars", @options[:toolbars]||[])
      }
    }

    # Search the index.
    #@editor_dlg.on("look_up") { |dlg, search_string|
    #  results = Index.instance.look_up(search_string, 1000).map{ |hash| hash[:id] }
    #  dlg.return results
    #}

    # Save a new snippet.
    @editor_dlg.on("save_snippet") { |dlg, id, code, metadata|
      success = save_snippet(id, code, metadata)
      dlg.return success
    }

    # Delete a new snippet.
    @editor_dlg.on("delete_snippet") { |dlg, id|
      delete_snippet(id)
    }

    # Save the toolbars and close the dialog.
    @editor_dlg.on("save_toolbars") { |dlg, array|
      dlg.close
      # Update toolbars as much as possible (we can't change order or remove buttons).
      begin
        array.each{ |hash|
          # Create new toolbars.
          if !@toolbars.include?(hash["id"])
            load_toolbar(hash)
          # Add new buttons to existing toolbars.
          else
            toolbar = @toolbars[hash["id"]]
            toolbar_commands = [] # Get all commands that this toolbar already has, ignore strings.
            toolbar.each{ |command| toolbar_commands << command if command.is_a?(UI::Command) }
            # Check if there are new buttons.
            if toolbar_commands.length < hash["buttons"].length
              hash["buttons"].each{ |id|
                next toolbar.add_separator() if id == 0
                entry = Index.instance.get_by_id(id)
                entry = @snippets.find{ |e| e[:id] == id } if entry.nil?
                # TODO: we get different Command objects for the same commands, so the following comparison failed:
                next if entry.nil? || toolbar_commands.find{ |c| c == entry[:command] || c.menu_text == entry[:name] }
                command = entry[:command] || UI::Command.new(entry[:name], &entry[:proc]) rescue next
                toolbar.add_item(command)
              }
            end
          end
        }
      ensure
        # Save the data.
        @options[:toolbars] = array
        @options.save
      end
    }

    # File dialog.
    @editor_dlg.on("openpanel") { |dlg|
      dlg.return File.expand_path(UI.openpanel)
    }

    # Close the dialog.
    @editor_dlg.on("close") { |dlg|
      dlg.close
    }

    # Show the webdialog.
    @editor_dlg.show

  end
  return @editor_dlg
end


# Creates toolbar objects for commands remembered from @options and identified by IDs.
def self.load_toolbars
  # Load custom buttons.
  @options[:snippets].each{ |hash|
    next if !hash.is_a?(Hash) || !hash["name"].is_a?(String)
    load_snippet(hash)
  }
  # Load toolbars.
  @options[:toolbars].each{ |hash|
    next if !hash.is_a?(Hash) || !hash["buttons"].is_a?(Array) || hash["buttons"].empty?
    load_toolbar(hash)
  }
end


# Creates a toolbar object from data in a hash.
# @param [Hash] hash - A hash containing 
#   "name" => [String],
#   "buttons" => [Array<Fixnum>]
def self.load_toolbar(hash)
  return puts("Argument hash must include a string 'name' and a non-empty array 'buttons'.") unless
      hash["name"].is_a?(String) &&
      hash["buttons"].is_a?(Array) &&
      hash["buttons"].length > 0
  # Collect all commands (to check that > 0 are found)
  commands = hash["buttons"].map{ |id|
    next 0 if id == 0 # separator
    entry = Index.instance.get_by_id(id)
    entry = @snippets.find{ |e| e[:id] == id } if entry.nil?
    next if entry.nil?
    command = entry[:command] || UI::Command.new(entry[:name], &entry[:proc]) rescue next
  }.compact
  # Avoid empty toolbars, they would show no title bar and only a millimeter wide.
  return unless commands.length > 0
  # Create a new toolbar and add the commands.
  toolbar = UI::Toolbar.new(hash["name"])
  commands.each{ |command|
    next toolbar.add_separator() if command == 0
    toolbar.add_item(command)
  }
  # Show toolbar if it was open when we shutdown.
  toolbar.restore
  # Per bug 2902434, adding a timer call to restore the toolbar. This
  # fixes a toolbar resizing regression on Windows as the restore() call
  # does not seem to work as the script is first loading.
  UI.start_timer(0.1, false) {
    toolbar.restore
  }
  @toolbars[hash["id"]] = toolbar
  return toolbar
end
private_class_method :load_toolbar


# Loads a snippet from a file and creates a command for it.
# @param [Hash] hash
def self.load_snippet(hash)
  return nil unless hash["id"].is_a?(Fixnum)
  id = hash["id"]
  snippet_path = File.join(DATA_DIR, "#{id}.rb")
  return nil unless File.exists?(snippet_path)
  code = IO.read(snippet_path)
  create_snippet_command(id, code, hash)
end
private_class_method :load_snippet


# Save a snippet to a file.
# @param [Fixnum] id
# @param [String] code
# @param [Hash] metadata
# @returns [Boolean] whether the snippet was saved successfully
def self.save_snippet(id, code, metadata={})
  return false unless id.is_a?(Fixnum) && code.is_a?(String)
  success = create_snippet_command(id, code, metadata)
  return false unless success
  Dir.mkdir(DATA_DIR) unless File.exists?(DATA_DIR)
  snippet_path = File.join(DATA_DIR, "#{id}.rb")
  File.open(snippet_path, "w") { |file|
    file.write(code)
  }
  metadata["id"] = id
  @options[:snippets] << metadata
  return true
rescue Errno::ENOENT => e
  e.message << " Could not write #{snippet_path}"
  if defined?(AE::Console)
    AE::Console.error(e)
  else
    $stderr.write(e.message << $/)
    $stderr.write(e.backtrace.join($/) << $/)
  end
  return false
end
private_class_method :save_snippet


# Deletes a snippet.
# @param [Fixnum] id
def self.delete_snippet(id)
  return nil unless id.is_a?(Fixnum)
  @options[:snippets].reject!{ |e| e[:id] == id }
  @snippets.reject!{ |e| e[:id] == id }
  snippet_path = File.join(DATA_DIR, "#{id}.rb")
  File.delete(snippet_path) rescue nil
end
private_class_method :delete_snippet


# Creates the command object of a snippet.
# Since we can currently neither set nor read the Proc of a UI::Command, it 
# creates a new command instance.
# @param [Fixnum] id
# @param [String] name
# @param [String] code
# @param [Hash] metadata
# @returns [Boolean] whether the command was created successfully
def self.create_snippet_command(id, code, metadata)
  name = metadata["name"]
  return false unless id.is_a?(Fixnum) && code.is_a?(String) && name.is_a?(String)
  # Create a Proc.
  block = eval(
    "Proc.new{
      begin
        #{code}
      rescue Exception => e
        e.message << ' in snippet #{name}'
        if defined?(AE::Console)
          AE::Console.error(e)
        else
          $stderr.write(e.message << $/)
          $stderr.write(e.backtrace.join($/) << $/)
        end
      end
    }"
  )
  raise unless block.is_a?(Proc)
  # Create a command.
  command = UI::Command.new(name, &block)
  command.tooltip = metadata["description"] || ""
  command.status_bar_text = metadata["instruction"] || ""
  command.large_icon = metadata["large_icon"] || metadata["icon"] || ""
  command.small_icon = metadata["small_icon"] || metadata["icon"] || ""
  # Create or update the entry in the snippet index.
  entry = @snippets.find{ |e| e[:id] == id} || {}
  entry[:id] ||= id
  entry[:name] = name
  entry[:category] = metadata["category"] if metadata.include?("category")
  entry[:description] = metadata["description"] if metadata.include?("description")
  entry[:icon] = metadata["icon"] if metadata.include?("icon")
  entry[:command] = command
  entry[:code] = code
  entry[:proc] = block
  @snippets << entry
  return true
rescue Exception => e
  e.message << " in snippet #{name}"
  if defined?(AE::Console)
    AE::Console.error(e)
  else
    $stderr.write(e.message << $/)
    $stderr.write(e.backtrace.join($/) << $/)
  end
  return false
end
private_class_method :create_snippet_command


=begin
# Switches a list of toolbars on, and all other (currently Ruby) toolbars off.
# @param [String] name - the name of the workspace.
# TODO: Needs a UI to create workspaces and assign toolbars to them.
def self.switch_to_workspace(name)
  @options[:workspaces].each{ |workspace|
    if workspace[:name] != name
      workspace[:toolbars].each{ |toolbar_id|
        @toolbars[toolbar_id].hide
      }
    else # workspace[:name] == name
      workspace[:toolbars].each{ |toolbar_id|
        @toolbars[toolbar_id].show
      }
    end
  }
end
=end


unless file_loaded?(File.basename(__FILE__))

  # Add this plugin to the UI
  cmd_editor = UI::Command.new(TRANSLATE["Toolbar Editor"]){ AE::ToolbarEditor.show_dialog }
  cmd_editor.tooltip = TRANSLATE["A drag&drop editor to create custom toolbars."]
  cmd_editor.small_icon = File.join(DIR, "images", "icon_toolbareditor_16.png")
  cmd_editor.large_icon = File.join(DIR, "images", "icon_toolbareditor_24.png")
  cmd_editor.set_validation_proc {
    (@editor_dlg && @editor_dlg.visible?) ? MF_CHECKED : MF_UNCHECKED
  }
  UI.menu("Window").add_item(cmd_editor)

  # Load all custom toolbars when SketchUp starts, but only after all other plugins have been loaded.
  # Alternative 1: require all in load path:
=begin
  $LOAD_PATH.each{ |path|
    next unless path.is_a?(String) && File.directory?(path)
    # The following freezes SketchUp:
    Dir[File.join(path, "*.{rb,rbs}")][0,10].each{ |file|
      next if file == __FILE__ || $".include?(file)
      begin
        require(file)
      rescue
        next
      end
    }
  }
  self.load_toolbars()
=end
  # Alternative 2: use timer
  # Problem: SketchUp2013 doesn't remember UI::Toolbar.get_last_state for toolbars
  # that were created long after loading finished. This was fixed in 2013.
  UI.start_timer(5, false) { self.load_toolbars() }

  file_loaded(File.basename(__FILE__))
end # end unless


end # module ToolbarEditor


end # module AE
