/*
AE JavaScript Library for SketchUp WebDialogs (extract)
Version:      1.1.4
Date:         02.04.2014
Summary:
module AE
  module Bridge                    // Communication with Ruby. Requires AE::Dialog
    .callRuby(name, *arguments [,callbackFunction]) // Executes a Ruby action_callback with any amount of arguments.
    .nextMessage()                 // Called from Ruby to send the next message from the queue.
    .callbackJS(id [,data])        // Called from a Ruby action_callback to run another JavaScript callback.
    .initialize()                  // Tell SketchUp that the dialog is ready.
*/


/**
 * @namespace
 */
var AE = function(AE) {


/**
 * Debugging mode
 */
AE.debug = false;


/**
 * Whether debugging is done in an external browser.
 * Disables the use of the skp protocol and logs to the browser's console.
 */
AE.debug_no_skp = false;


/**
 * @const PLATFORM
 */
AE.PLATFORM = (navigator.appVersion.indexOf("Win") != -1)? "WIN" : ((navigator.appVersion.indexOf("OS X")!=-1)? "OSX" : "other");
AE.RUBY_VERSION = 0;     // Needs to be set from SketchUp.
AE.SKETCHUP_VERSION = 0; // Needs to be set from SketchUp.


/**
 * @module Translate
 * Stub module for the case that no translation has been loaded.
 */
if (!AE.Translate) { AE.Translate = { get: function(s) { return s; } }; }


/**
 * @module Bridge:
 * Methods to interact with SketchUp/Ruby.
 */
AE.Bridge = function(self) {


var messageID = 0;

// Private object to hold callback functions.
var callbacks = {};

// SketchUp/OSX/Safari skips skp urls if they happen in a too short time interval.
// We pass all skp urls through a queue that makes sure that a message is only
// sent after the SketchUp side has received the previous message.
var messageQueue = [];
var messageField = null;

// Whether the queue is ready to send another messge.
var ready = true;


/**
 * Calls a SketchUp Ruby callback.
 * @param {string} name - The name of the Ruby action_callback.
 * @param {...Object} argument - Any amount of arguments of JSON type.
 * @param {function} callback - A JavaScript function to execute after the action_callback.
 * @returns {number} - The id of the sent message.
 */
self.callRuby = function(name, argument, callback) {
  var args = Array.prototype.slice.call(arguments).slice(1),
    callback = (typeof(args[args.length-1]) === 'function') ? args.pop() : null,
    data = {id: messageID, name: name, arguments: args};
  // In case of debugging, just log the arguments to the console.
  if (AE.debug_no_skp) {
    console.log(name, args);
  } else {
    data = (AE.RUBY_VERSION > 1.9) ? JSON.stringify(data) : JSONstringify(data);
    // We assign an id to this message so we can identify a callback (if there is one).
    if (callback) { callbacks[messageID] = callback; }
    var url = "skp:AE.Bridge.receive_message";
    messageQueue.push([url, data]);
    // If the queue is not running, start it. If the message queue contains many urls, then it is already running.
    if (ready) { self.nextMessage(); }
  }
  // Increase the id for the next message.
  messageID++;
  // Return the id of this message.
  return messageID - 1;
};


/**
 * Tells the bridge that the next message can be sent.
 * @private (only for use by this file or Ruby AE::Dialog)
 */
self.nextMessage = function() {
  var message = messageQueue.shift();
  if (!message) { ready = true; return; }
  // Lock the status variable before sending the message.
  // (because window.location is synchronous in IE and finishes before this
  // function finishes.)
  ready = false;
  // Prepare the message field, so that SketchUp can obtain the message data from here.
  var url = message[0],
      data = message[1];
  fillMessageField(data);
  // Give enough time to refresh the DOM, so that SketchUp will be able to
  // access the latest values of input elements.
  window.setTimeout(function(){
    window.location.href = url;
  }, 0);
};


/**
 * Calls a JavaScript callback from a SketchUp Ruby action_callback.
 * @private (only for use by Ruby AE::Dialog)
 * @param {number} id - The id of the JavaScript-to-Ruby message.
 * @param {...argument} argument - Any amount of arguments.
 */
self.callbackJS = function(id, argument) {
  var args = Array.prototype.slice.call(arguments).slice(1);
  // If there is a callback, execute it.
  if (id && callbacks[id]) {
    try {
      callbacks[id].apply(window, args)
    } catch (e) {
      self.error("AE.Bridge.callbackJS: Error when executing callback #"+id+": "+e);
    }
    delete callbacks[id];
  }
};


/**
 * Tells SketchUp that the dialog is loaded.
 * AE::Dialog will run on_show callbacks.
 */
self.initialize = function() {
  self.callRuby("AE.Bridge.initialize");
};


/* For debugging. */
self.puts = function(text) {
  self.callRuby("puts",  text);
};


/* For debugging. */
self.error = function(text) {
  self.callRuby("error",  text);
};


/* For debugging. If no browser console defined, link to the Ruby Console. */
window.console = window.console || {
  log: function(text) { self.puts(text); }
};


/** private */


function fillMessageField(value) {
  if (!messageField) { createMessageField(); }
  messageField.value = value;
};


function createMessageField() {
  messageField = document.createElement('input');
  messageField.setAttribute('type', 'hidden');
  messageField.setAttribute('style', 'display: none');
  messageField.setAttribute('id', 'AE.Bridge.messageField');
  document.getElementsByTagName("body")[0].appendChild(messageField);
};


/**
 * Fallback JSON.stringify method that does not encode unicode characters.
 * Important for older Ruby versions that don't understand unicode escaping (\u00f6)
 * but treat unicode multibyte characters as multiple single byte characters.
 * @param {object} object
 * @returns {string}
 */
function JSONstringify(object) {
  var jsonString = '';
  // null
  if (object === null || typeof object === 'undefined') {
    jsonString = 'null';
  // array
  } else if (Object.prototype.toString.call(object) === '[object Array]') {
    var a = [];
    for (var i = 0; i < object.length; i++) {
      var v = JSONstringify(object[i]);
      if (!!v) { a.push(v) }
    }
    jsonString = '[' + a.join(', ') + ']';
  // object literal
  } else if (Object.prototype.toString.call(object) === '[object Object]') {
    var o = [];
    for (var e in object) {
      if (e === object) { continue; } // ignore circle references. Should never occur!
      // if (Object.prototype.hasOwnProperty.call(object, e)) { // causes "out of memory" ?
      var k = JSONstringify(e);
      var v = JSONstringify(object[e]);
      if (!!k && !!v) { o.push(k + ': ' + v) }
      // }
    }
    jsonString = '{' + o.join(', ') + '}';
  // string
  } else if (typeof object === 'string') {
    jsonString = '"' + object.replace(/\\/g, '\\\\').replace(/\//g, '\\/')
      .replace(/\"/g, '\\\"').replace(/\f/g, '\\f')
      .replace(/\n/g, '\\n').replace(/\r/g, '\\r').replace(/\t/g, '\\t') + '"';
  // number
  } else if (typeof object === 'number') {
    if (isNaN(object)) { jsonString = '0.0/0.0'; }
    else if (!isFinite(object)) { jsonString = (object > 0) ? '1.0/0.0' : '-1.0/0.0'; }
    else { jsonString = String(object); }
  } else if (typeof object === 'boolean') {
    jsonString = String(object);
  } // else: undefined
  return jsonString;
}


return self;
}(AE.Bridge || {}); // end @module Bridge


return AE;
}(AE || {}); // end @module AE
